# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from .inspector import Inspector


# Module API

def validate(source, **options):
    """Validates a source file and returns a report.

    # Arguments

        source (Union[str, Dict, List[Dict], IO]):
            The source to be validated.
            It can be a local file path, URL, dict, list of dicts, or a
            file-like object. If it's a list of dicts and the `preset` is
            "nested", each of the dict key's will be used as if it was passed
            as a keyword argument to this method.

            The file can be a CSV, XLS, JSON, and any other format supported by
            `tabulator`_.

        checks (List[str]):
            List of checks names to be enabled. They can be
            individual check names (e.g. `blank-headers`), or check types (e.g.
            `structure`).

        skip_checks (List[str]):
            List of checks names to be skipped. They can
            be individual check names (e.g. `blank-headers`), or check types
            (e.g.  `structure`).

        infer_schema (bool):
            Infer schema if one wasn't passed as an argument.

        infer_fields (bool):
            Infer schema for columns not present in the received schema.

        order_fields (bool):
            Order source columns based on schema fields order.
            This is useful when you don't want to validate that the data
            columns' order is the same as the schema's.

        error_limit (int):
            Stop validation if the number of errors per table exceeds this value.

        table_limit (int):
            Maximum number of tables to validate.

        row_limit (int):
            Maximum number of rows to validate.

        preset (str):
            Dataset type could be `table` (default), `datapackage`,
            `nested` or custom. Usually, the preset can be inferred from the
            source, so you don't need to define it.

        Any (Any):
            Any additional arguments not defined here will be passed on,
            depending on the chosen `preset`. If the `preset` is `table`, the
            extra arguments will be passed on to `tabulator`_, if it is
            `datapackage`, they will be passed on to the `datapackage`_
            constructor.

    # Raises
        GoodtablesException: Raised on any non-tabular error.

    # Returns
        dict: The validation report.

    """
    source, options, inspector_settings = _parse_arguments(source, **options)

    # Validate
    inspector = Inspector(**inspector_settings)
    report = inspector.inspect(source, **options)

    return report


def _parse_arguments(source, **options):
    # Extract settings
    validation_options = set((
        'checks',
        'skip_checks',
        'infer_schema',
        'infer_fields',
        'order_fields',
        'error_limit',
        'table_limit',
        'row_limit',
        'custom_presets',
        'custom_checks',
    ))
    settings = dict((
        (key, options.pop(key)) for key in validation_options
        if key in options
    ))

    # Support for pathlib.Path
    if hasattr(source, 'joinpath'):
        source = str(source)
    if isinstance(source, list):
        if source and isinstance(source[0], dict) and 'source' in source[0]:
            for index, item in enumerate(source):
                if hasattr(item['source'], 'joinpath'):
                    source[index]['source'] = str(item['source'])

    return source, options, settings



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import pytest
from copy import deepcopy
from goodtables import validate
from goodtables.presets.datapackage import datapackage


# Validate

@pytest.mark.parametrize('dp_path', [
    'data/datapackages/valid/datapackage.json',
    'data/datapackages/valid.zip',
])
def test_validate_datapackage_valid(log, dp_path):
    report = validate(dp_path)
    assert log(report) == []


@pytest.mark.parametrize('dp_path', [
    'data/datapackages/invalid/datapackage.json',
    'data/datapackages/invalid.zip',
])
def test_validate_datapackage_invalid(log, dp_path):
    report = validate(dp_path)
    assert log(report) == [
        (1, 3, None, 'blank-row'),
        (2, 4, None, 'blank-row'),
    ]


# Validate (integrity)

DESCRIPTOR = {
    'resources': [
        {
            'name': 'resource1',
            'path': 'data/valid.csv',
            'bytes': 30,
            'hash': 'sha256:a1fd6c5ff3494f697874deeb07f69f8667e903dd94a7bc062dd57550cea26da8',
        }
    ]
}

def test_check_file_integrity(log):
    source = deepcopy(DESCRIPTOR)
    report = validate(source)
    assert log(report) == []


def test_check_file_integrity_invalid(log):
    source = deepcopy(DESCRIPTOR)
    source['resources'][0]['bytes'] += 1
    source['resources'][0]['hash'] += 'a'
    report = validate(source)
    assert report['tables'][0]['errors'] == [{
        'code': 'source-error',
        'message': 'Calculated size "30" and hash "a1fd6c5ff3494f697874deeb07f69f8667e903dd94a7bc062dd57550cea26da8" differ(s) from declared value(s)',
        'message-data': {}
    }]


def test_check_file_integrity_size(log):
    source = deepcopy(DESCRIPTOR)
    source['resources'][0]['hash'] = None
    report = validate(source)
    assert log(report) == []


def test_check_file_integrity_size_invalid(log):
    source = deepcopy(DESCRIPTOR)
    source['resources'][0]['bytes'] += 1
    source['resources'][0]['hash'] = None
    report = validate(source)
    assert report['tables'][0]['errors'] == [{
        'code': 'source-error',
        'message': 'Calculated size "30" differ(s) from declared value(s)',
        'message-data': {}
    }]


def test_check_file_integrity_hash(log):
    source = deepcopy(DESCRIPTOR)
    source['resources'][0]['bytes'] = None
    report = validate(source)
    assert log(report) == []


def test_check_file_integrity_hash_invalid(log):
    source = deepcopy(DESCRIPTOR)
    source['resources'][0]['bytes'] = None
    source['resources'][0]['hash'] += 'a'
    report = validate(source)
    assert report['tables'][0]['errors'] == [{
        'code': 'source-error',
        'message': 'Calculated hash "a1fd6c5ff3494f697874deeb07f69f8667e903dd94a7bc062dd57550cea26da8" differ(s) from declared value(s)',
        'message-data': {}
    }]


def test_check_file_integrity_invalid(log):
    source = deepcopy(DESCRIPTOR)
    source['resources'][0]['hash'] = 'not-supported-hash'
    report = validate(source)
    assert report['warnings'] == [
        'Resource "resource1" does not use the SHA256 hash. The check will be skipped',
    ]


# Preset

def test_preset_datapackage():
    warnings, tables = datapackage('data/datapackages/valid/datapackage.json')
    assert len(warnings) == 0
    assert len(tables) == 2


# Issues

def test_preset_datapackage_non_tabular_datapackage_issue_170():
    warnings, tables = datapackage('data/non_tabular_datapackage.json')
    assert len(warnings) == 0
    assert len(tables) == 2


def test_preset_datapackage_mixed_datapackage_issue_170():
    warnings, tables = datapackage('data/mixed_datapackage.json')
    assert len(warnings) == 0
    assert len(tables) == 2


def test_preset_datapackage_invalid_json_issue_192():
    warnings, tables = datapackage('data/invalid_json.json')
    assert len(warnings) == 1
    assert len(tables) == 0
    assert 'Unable to parse JSON' in warnings[0]



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import pytest
from goodtables import validate
from goodtables.presets.table import table
from goodtables.presets.datapackage import datapackage
from goodtables.presets.nested import nested


# Validate

def test_validate_tables_invalid(log):
    report = validate([
        {'source': 'data/valid.csv',
         'schema': {'fields': [{'name': 'id'}, {'name': 'name'}]}},
        {'source': 'data/invalid.csv'},
    ], preset='nested', infer_schema=True)
    assert log(report) == [
        (2, None, 3, 'blank-header'),
        (2, None, 4, 'duplicate-header'),
        (2, 2, 3, 'missing-value'),
        (2, 2, 4, 'missing-value'),
        (2, 3, None, 'duplicate-row'),
        (2, 4, None, 'blank-row'),
        (2, 5, 5, 'extra-value'),
    ]


def test_validate_nested_presets_set_default_preset():
    report = validate([
        {'source': 'data/datapackages/valid/datapackage.json'},
    ], preset='nested', infer_schema=True)
    assert report['valid']
    assert report['warnings'] == []


# Preset

def test_preset_nested():
    warnings, tables = nested([
        {'source': 'data/valid.csv'},
        {'source': 'data/invalid.csv', 'preset': 'table'},
        {'source': 'data/datapackages/valid/datapackage.json', 'preset': 'datapackage'},
        {'source': 'data/datapackages/invalid/datapackage.json', 'preset': 'datapackage'},
    ], presets={
        'table': {'func': table},
        'nested': {'func': nested},
        'datapackage': {'func': datapackage},
    })
    assert len(warnings) == 0
    assert len(tables) == 6



from datapackage import Package


# Module API

def init_datapackage(resource_paths):
    # "Create tabular data package with resources.
    dp = Package({
        'name': 'change-me',
        'schema': 'tabular-data-package',
    })

    for path in resource_paths:
        dp.infer(path)

    return dp



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import pytest
from goodtables import validate
from goodtables.presets.table import table


# Validate

def test_validate_table_valid(log):
    report = validate('data/valid.csv')
    assert log(report) == []


def test_validate_table_invalid(log):
    report = validate('data/invalid.csv', infer_schema=True)
    assert log(report) == [
        (1, None, 3, 'blank-header'),
        (1, None, 4, 'duplicate-header'),
        (1, 2, 3, 'missing-value'),
        (1, 2, 4, 'missing-value'),
        (1, 3, None, 'duplicate-row'),
        (1, 4, None, 'blank-row'),
        (1, 5, 5, 'extra-value'),
    ]


def test_validate_table_invalid_error_limit(log):
    report = validate('data/invalid.csv', error_limit=2, infer_schema=True)
    assert log(report) == [
        (1, None, 3, 'blank-header'),
        (1, None, 4, 'duplicate-header'),
    ]


def test_validate_table_invalid_row_limit(log):
    report = validate('data/invalid.csv', row_limit=2, infer_schema=True)
    assert log(report) == [
        (1, None, 3, 'blank-header'),
        (1, None, 4, 'duplicate-header'),
        (1, 2, 3, 'missing-value'),
        (1, 2, 4, 'missing-value'),
    ]

# Preset

def test_preset_table():
    warnings, tables = table('data/valid.csv')
    assert len(warnings) == 0
    assert len(tables) == 1


def test_preset_table_but_got_datapackage_issue_187():
    warnings, tables = table('data/datapackages/valid/datapackage.json')
    assert len(warnings) == 1
    assert len(tables) == 0
    assert 'Use "datapackage" preset' in warnings[0]


def test_preset_table_invalid_json_issue_196():
    warnings, tables = table('valid.csv', schema='data/invalid_json.json')
    assert len(warnings) == 1
    assert len(tables) == 0
    assert 'has a loading error' in warnings[0]



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from copy import deepcopy
from goodtables import validate


# Validate

FK_DESCRIPTOR = {
  'resources': [
    {
      'name': 'cities',
      'data': [
        ['id', 'name', 'next_id'],
        [1, 'london', 2],
        [2, 'paris', 3],
        [3, 'rome', 4],
        [4, 'rio', None],
      ],
      'schema': {
        'fields': [
          {'name': 'id', 'type': 'integer'},
          {'name': 'name', 'type': 'string'},
          {'name': 'next_id', 'type': 'integer'},
        ],
        'foreignKeys': [
          {
            'fields': 'next_id',
            'reference': {'resource': '', 'fields': 'id'},
          },
          {
            'fields': 'id',
            'reference': {'resource': 'people', 'fields': 'label'},
          },
        ],
      },
    }, {
      'name': 'people',
      'data': [
        ['label', 'population'],
        [1, 8],
        [2, 2],
        [3, 3],
        [4, 6],
      ],
    },
  ],
}


def test_foreign_key(log):
    descriptor = deepcopy(FK_DESCRIPTOR)
    report = validate(descriptor, checks=['foreign-key'])
    assert log(report) == []


def test_foreign_key_not_defined_foreign_keys(log):
    descriptor = deepcopy(FK_DESCRIPTOR)
    del descriptor['resources'][0]['schema']['foreignKeys']
    report = validate(descriptor, checks=['foreign-key'])
    assert log(report) == []


def test_foreign_key_source_is_not_datapackage(log):
    report = validate('data/valid.csv', checks=['foreign-key'])
    assert log(report) == []


def test_foreign_key_self_referenced_resource_violation(log):
    descriptor = deepcopy(FK_DESCRIPTOR)
    del descriptor['resources'][0]['data'][4]
    report = validate(descriptor, checks=['foreign-key'])
    assert log(report) == [
        (1, 4, None, 'foreign-key'),
    ]


def test_foreign_key_internal_resource_violation(log):
    descriptor = deepcopy(FK_DESCRIPTOR)
    del descriptor['resources'][1]['data'][4]
    report = validate(descriptor, checks=['foreign-key'])
    assert log(report) == [
        (1, 5, None, 'foreign-key'),
    ]


def test_foreign_key_internal_resource_violation_non_existent(log):
    descriptor = deepcopy(FK_DESCRIPTOR)
    del descriptor['resources'][1]
    report = validate(descriptor, checks=['foreign-key'])
    assert log(report) == [
        (1, 2, None, 'foreign-key'),
        (1, 3, None, 'foreign-key'),
        (1, 4, None, 'foreign-key'),
        (1, 5, None, 'foreign-key'),
    ]


def test_foreign_key_external_resource(log):
    descriptor = 'data/datapackages_linked/cities/datapackage.json'
    report = validate(descriptor, checks=['structure', 'schema', 'foreign-key'])
    assert log(report) == []


def test_foreign_key_external_resource_errors(log):
    descriptor = 'data/datapackages_linked_errors/cities/datapackage.json'
    report = validate(descriptor, checks=['structure', 'schema', 'foreign-key'])
    assert log(report) == [
        (1, 4, None, 'foreign-key'),  # self-referenced
        (1, 4, None, 'foreign-key'),  # external
    ]


def test_foreign_key_external_resource_remote_datapackage(log):
    descriptor = {
        'resources': [{
          'name': 'countries',
          'data': [
            ['Country Code', 'Country Name'],
            ['PRT', 'Portugal'],
            ['DRL', 'Dreamland'],
          ],
          'schema': {
            'fields': [
              {'name': 'Country Code', 'type': 'string'},
              {'name': 'Country Name', 'type': 'string'},
            ],
            'foreignKeys': [
              {
                'fields': 'Country Code',
                'reference': {'package': 'https://raw.githubusercontent.com/datasets/gdp/master/datapackage.json', 'resource': 'gdp', 'fields': 'Country Code'},
              },
            ],
          },
          }]
    }
    report = validate(descriptor, checks=['structure', 'schema', 'foreign-key'])
    assert log(report) == [
        (1, 3, None, 'foreign-key'),
    ]



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from copy import copy
from ..registry import check
from ..error import Error


# Module API

@check('missing-value')
def missing_value(cells):
    """
    missing-value: 	A row has less columns than the header.
    """
    errors = []

    for cell in copy(cells):

        # Skip if cell has value
        # There is a difference between:
        # - not having value at all - there is no `value` key
        # - having a value which is falsy (None, False, '', etc)
        # (so we don't use something like `if cell.get('value')`)
        if 'value' in cell:
            continue

        # Add error
        error = Error('missing-value', cell)
        errors.append(error)

        # Remove cell
        cells.remove(cell)

    return errors



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals
from . import config
__version__ = config.VERSION


# Module API

# TODO: somethin wrong goes on with cli here
# The same pattern with `__main__/cli` works fine for tabulator/tableschema/datapackage
# There we don't need to have `__name__ == "__main__":` in cli.py
# And following line doesn't lead to a warning
# from .cli import cli
from .validate import validate
from .registry import preset
from .registry import check
from .error import Error
from .spec import spec
from .exceptions import GoodtablesException

# Deprecated

from . import (exceptions,)
from .helpers import (init_datapackage,)
from .inspector import (Inspector,)

# Register

import importlib
from . import config
for module in config.PRESETS:
    importlib.import_module(module)
for module in config.CHECKS:
    importlib.import_module(module)
